@@ -23,15 +23,10 @@
 #include "stat.h"       /* for start_song_picker() */
 #include "player.h"     /* for player_queue_command() */
 #include "display.h"    /* for draw_info() and draw_screen() */
+#include "play_ogg.h"   /* for ogg_load_meta() */
+#include "play_mp3.h"   /* for mp3_load_meta() */
 #include "database.h"
 
-/* File Extensions to use. */
-const db_extension_info_t db_extensions[] = {
-    {"info", BASENAME_META, insert_meta_data, NULL, NULL },
-    {"stat", BASENAME_STAT, set_stat_data, save_stat_data, is_stat_data_changed}
-};
-const int db_extensions_size = sizeof( db_extensions ) / sizeof( db_extensions[0] );
-
 /*
  * This is called by main() to load the DB.  Since nothing else can happen until
  * the database is loaded, it does a few other things, in order to tell the other
@@ -45,15 +40,12 @@ void *setup_database( void *data ) {
     squash_lock( player_info.lock );
 
     /* Load the database */
-    load_db();
+    load_db_filenames();
 
     /* Unlock the locks we don't need anymore */
     squash_unlock( player_info.lock );
     squash_unlock( song_queue.lock );
 
-    /* Load the statistics routine (needed for playlist_manager() to call pick_song()) */
-    start_song_picker();
-
     /* Hide the info window, and unhide the spectrum analyzer */
     display_info.window[ WIN_INFO ].state = WIN_STATE_HIDDEN;
     display_info.window[ WIN_SPECTRUM ].state = WIN_STATE_NORMAL;
@@ -83,21 +75,32 @@ void *setup_database( void *data ) {
     squash_signal( song_queue.not_full );
     squash_unlock( song_queue.lock );
 
+    load_all_meta_data( TYPE_STAT ); /* Load statistics */
+    /* Load the statistics routine (needed for playlist_manager() to call pick_song()) */
+    start_song_picker();
+    squash_signal( database_info.stats_finished );
+
+    load_all_meta_data( TYPE_META ); /* Load info files */
+
     return (void *)NULL;
 }
 
 /*
  * Loads the database.  Will walk the config.db_paths[ BASENAME_SONG ] directory
- * looking for known music types.  It will then load the meta and stat values too.
- * (which may be in a different directory tree).
+ * looking for known music types.  Loading the meta data and stat values are done,
+ * at a later time, this only loads filenames.
  */
-void load_db( void ) {
+void load_db_filenames( void ) {
     struct stat path_stat;
 
     /* Initialize database_info */
-    database_info.song_count = 0;
-    database_info.song_count_allocated = 0;
-    database_info.songs = NULL;
+    if( database_info.songs != NULL ) {
+        squash_free( database_info.songs );
+        database_info.song_count = 0;
+        database_info.song_count_allocated = 0;
+        database_info.songs = NULL;
+        database_info.stats_loaded = 0;
+    }
 
     /* Check the path to see if it exists */
     if( stat(config.db_paths[ BASENAME_SONG ], &path_stat) != 0 ) {
@@ -279,8 +282,8 @@ void clear_db() {
  * really does any work for the ".stat" files; not the ".info" files
  */
 void save_song( song_info_t *song ) {
-    int cur_file_length;
     char *cur_filename;
+    char *end, *dirname;
     FILE *cur_file;
     int i;
 
@@ -298,10 +301,16 @@ void save_song( song_info_t *song ) {
         }
 
         /* Construct a filename to save */
-        cur_file_length = strlen(song->basename[db_extensions[i].which_basename]) + strlen(song->filename) + strlen(db_extensions[i].extension) + 3;
-        squash_malloc( cur_filename, cur_file_length );
+        cur_filename = build_fullfilename( song, db_extensions[i].which_basename );
 
-        snprintf( cur_filename, cur_file_length, "%s/%s.%s", song->basename[db_extensions[i].which_basename], song->filename, db_extensions[i].extension);
+        /* find dir part */
+        end = strrchr( cur_filename, '/' );
+        if( end != NULL ) {
+            dirname = copy_string( cur_filename, end - 1 );
+            /* ensure path is created */
+            create_path( dirname );
+            squash_free( dirname );
+        }
 
         /* If you can't open the file bomb */
         if( (cur_file = fopen(cur_filename, "w")) == NULL ) {
@@ -341,40 +350,132 @@ void save_stat_data( song_info_t *song, FILE *file ) {
     fprintf( file, "play_count=%d\n", song->stat.play_count );
     fprintf( file, "skip_count=%d\n", song->stat.skip_count );
     fprintf( file, "repeat_counter=%d\n", song->stat.repeat_counter );
+    fprintf( file, "\n" );
 
     /* Reset the changed flag */
     song->stat.changed = FALSE;
 }
 
 /*
- * Loads the metadata for a song from the disk
+ * Save statistical info.  This is should be called by save_song()
  */
-void load_meta_data( song_info_t *song ) {
-    char *cur_file;
-    int cur_file_length;
-    int i;
+void save_meta_data( song_info_t *song, FILE *file ) {
+    int i, j;
 
-    /* Don't let them trick us! */
-    if( song == NULL || song->filename == NULL ) {
+    /* Some one is tricking us */
+    if( song == NULL || file == NULL ) {
         return;
     }
 
-    /* Parse each extension */
-    for( i = 0; i < db_extensions_size; i++ ) {
-        /* Determine the length of the filename to open */
-        cur_file_length = strlen(song->basename[db_extensions[i].which_basename]) + strlen(song->filename) + strlen(db_extensions[i].extension) + 3;
+    for(i = 0; i < song->meta_key_count; i++ ) {
+        for( j = 0; j < song->meta_keys[i].value_count; j++ ) {
+            fprintf( file, "%s=%s\n", song->meta_keys[i].key, song->meta_keys[i].values[j] );
+        }
+    }
+    fprintf( file, "\n" );
+}
 
-        /* Allocate memory to build the filename */
-        squash_malloc( cur_file, cur_file_length );
+/*
+ * Loads the metadata for a song from the disk
+ */
+void load_meta_data( song_info_t *song, enum meta_type_e which ) {
+    char *filename, *metaname, *dirname;
+    char *end;
+    enum song_type_e song_type;
+    bool success;
+    FILE *meta_file;
+    struct stat file_stat;
 
-        /* Build the filename */
-        snprintf( cur_file, cur_file_length, "%s/%s.%s", song->basename[db_extensions[i].which_basename], song->filename, db_extensions[i].extension);
+    /* Don't let them trick us! */
+    if( song == NULL ) {
+        return;
+    }
 
-        /* Parse the file */
-        parse_file( cur_file, db_extensions[i].add_data, (void *)song );
+    metaname = build_fullfilename( song, db_extensions[which].which_basename );
 
-        /* Free the file name */
-        squash_free( cur_file );
+    /* Parse the file unless we want to reload/save from the original song file */
+    if( config.db_overwriteinfo && config.db_saveinfo && which == TYPE_META ) {
+        success = false;
+    } else {
+        success = parse_file( metaname, db_extensions[which].add_data, (void *)song );
+    }
+
+    /* If we didn't load from the stat/info file, and we have a info file,
+     * try to load it from the original song file */
+    if( !success && which == TYPE_META ) {
+        song_type = get_song_type( song->filename );
+
+        if( song_type != TYPE_UNKNOWN ) {
+            filename = build_fullfilename( song, BASENAME_SONG );
+
+            if( song_type == TYPE_OGG ) {
+                ogg_load_meta( (void *)song, filename );
+            } else {
+                /* This routine does nothing right now */
+                mp3_load_meta( (void *)song, filename );
+            }
+            squash_free( filename );
+
+            if( config.db_saveinfo ) {
+#ifdef NO_VORBIS_COMMENT
+                if( song_type != TYPE_OGG ) {
+#else
+                if( TRUE ) {
+#endif
+#ifdef NO_ID3LIB
+                if( song_type != TYPE_MP3 ) {
+#else
+                if( TRUE ) {
+#endif
+                /* find dir part */
+                end = strrchr( metaname, '/' );
+                if( end != NULL ) {
+                    dirname = copy_string( metaname, end - 1 );
+                    /* ensure path is created */
+                    create_path( dirname );
+                    squash_free( dirname );
+                }
+
+                /* check to see if the file already exists and don't save unless
+                 * we are supposed to overwrite */
+                if( stat( metaname, &file_stat ) != 0 || config.db_overwriteinfo ) {
+                    /* If you can't open the file bomb */
+                    if( (meta_file = fopen(metaname, "w")) == NULL ) {
+                        squash_error( "Can't open file \"%s\" for writing", metaname );
+                    }
+                    save_meta_data( song, meta_file );
+                    fclose(meta_file);
+                }
+/* close two #ifdefs: */ }}
+            }
+        }
+    }
+
+    /* Free the file name */
+    squash_free( metaname );
+}
+
+/*
+ * Goes though all songs and loads the meta data for them.
+ */
+void load_all_meta_data( enum meta_type_e which ) {
+    int i;
+    for( i = 0; i < database_info.song_count; i++ ) {
+        squash_wlock( database_info.lock );
+        /* If the type is meta, we need to be careful */
+        if( which == TYPE_META ) {
+            /* if we haven't loaded before, set the key count to 0,
+             * otherwise, skip loading this song (since it has already
+             * been loaded */
+            if( database_info.songs[i].meta_key_count == -1 ) {
+                database_info.songs[i].meta_key_count = 0;
+            } else {
+                squash_wunlock( database_info.lock );
+                continue;
+            }
+        }
+        load_meta_data( &database_info.songs[i], which );
+        squash_wunlock( database_info.lock );
     }
 }
 
@@ -390,6 +491,10 @@ void insert_meta_data( void *data, char *header, char *key, char *value ) {
 
     /* Make sure we were given at least a song and a key */
     if( (song == NULL) || (key == NULL) ) {
+        /* shouldn't happen, but let's be safe */
+        if( value != NULL ) {
+            squash_free( value );
+        }
         return;
     }
 
@@ -412,6 +517,9 @@ void insert_meta_data( void *data, char *header, char *key, char *value ) {
         meta_key->key = key;
         meta_key->value_count = 0;
         meta_key->values = NULL;
+    } else {
+        /* Otherwise we already have a copy of key, so free this extra copy */
+        squash_free( key );
     }
 
     /* Increase size of space for the meta info */
@@ -432,6 +540,10 @@ void set_stat_data( void *data, char *header, char *key, char *value ) {
 
     /* Make sure we were given at least a song and a key */
     if( (song == NULL) || (key == NULL) ) {
+        /* shouldn't happen, but let's be safe */
+        if( value != NULL ) {
+            squash_free( value );
+        }
         return;
     }
 
@@ -446,6 +558,9 @@ void set_stat_data( void *data, char *header, char *key, char *value ) {
     } else if( strncasecmp("skip_count", key, 11) == 0 ) {
         song->stat.skip_count = int_value;
     }
+
+    squash_free( key );
+    squash_free( value );
 }
 
 /*
@@ -476,17 +591,14 @@ void _load_file( char *filename ) {
         song->basename[ BASENAME_META ] = config.db_paths[ BASENAME_META ];
         song->basename[ BASENAME_STAT ] = config.db_paths[ BASENAME_STAT ];
         song->meta_keys = NULL;
-        song->meta_key_count = 0;
+        song->meta_key_count = -1;
         song->stat.changed = FALSE;
 
-        /* Load the song from disk */
-        load_meta_data( song );
-
         /* Update the counter */
         database_info.song_count++;
 
-        /* Display progress every 100 songs */
-        if( database_info.song_count % 100 == 0) {
+        /* Display progress every 500 songs */
+        if( database_info.song_count % 500 == 0) {
             draw_info();
         }
     }
@@ -499,16 +611,21 @@ void _walk_filesystem( char *base_dir, void(*loader)(char *) ) {
     DIR *dir;
     struct dirent *file_entry;
     struct stat file_stat;
-    char *file_path;
-    char *fullfile_path;
-    char *fullbase_dir;
+    char *file_path = NULL;
+    char *fullfile_path = NULL;
+    char *fullbase_dir = NULL;
+    int file_entry_length;
+    int base_dir_length, fullbase_dir_length;
 
     /* Set the fullbase_dir */
     if( base_dir == NULL ) {
         fullbase_dir = strdup( config.db_paths[ BASENAME_SONG ] );
+        base_dir_length = 0;
     } else {
         squash_asprintf(fullbase_dir, "%s/%s", config.db_paths[ BASENAME_SONG ], base_dir );
+        base_dir_length = strlen( base_dir );
     }
+    fullbase_dir_length = strlen( fullbase_dir );
 
     /* Open the directory */
     if( (dir = opendir(fullbase_dir)) == NULL ) {
@@ -518,21 +635,32 @@ void _walk_filesystem( char *base_dir, void(*loader)(char *) ) {
 
     /* For each directory entry, recursively call ourselves if it's a directory
        otherwise call loader() if it's a file */
+    if( base_dir != NULL ) {
+        squash_malloc( file_path, base_dir_length + 257 );
+        strcpy( file_path, base_dir );
+        file_path[ base_dir_length ] = '/';
+        file_path[ base_dir_length+1 ] = '\0';
+    }
+    squash_malloc( fullfile_path, fullbase_dir_length + 257 );
+    strcpy( fullfile_path, fullbase_dir );
+    fullfile_path[ fullbase_dir_length ] = '/';
+    fullfile_path[ fullbase_dir_length+1 ] = '\0';
     while( (file_entry = readdir(dir)) != NULL ) {
         /* Ignore current/parent directory entries */
         if( (strncmp(".", file_entry->d_name, 2) == 0) ||
             (strncmp("..", file_entry->d_name, 3) == 0)) {
             continue;
         }
+        file_entry_length = strlen( file_entry->d_name );
 
         /* Build file_path */
         if( base_dir == NULL ) {
-            file_path = strdup( file_entry->d_name );
+            file_path = file_entry->d_name;
         } else {
-            squash_asprintf( file_path, "%s/%s", base_dir, file_entry->d_name );
+            strcpy( &file_path[ base_dir_length + 1 ], file_entry->d_name );
         }
         /* Bulid fullfile_path */
-        squash_asprintf( fullfile_path, "%s/%s", fullbase_dir, file_entry->d_name );
+        strcpy( &fullfile_path[ fullbase_dir_length + 1 ], file_entry->d_name );
 
         /* If the file can be stat'd, continue walking process */
         if( stat(fullfile_path, &file_stat) == 0 ) {
@@ -545,10 +673,12 @@ void _walk_filesystem( char *base_dir, void(*loader)(char *) ) {
             }
         }
 
-        /* Free built names */
-        squash_free( file_path );
-        squash_free( fullfile_path );
     }
+    /* Free built names */
+    if( base_dir != NULL ) {
+        squash_free( file_path );
+    }
+    squash_free( fullfile_path );
 
     /* Free full_basedir */
     squash_free( fullbase_dir );
